'use strict'

const Trait = require('../../entities/trait')
const Collection = require('../../entities/collection')

const combinatorFactory = require('./combinator')

function validateOptions(options) {
  const rarites = options?.rarities || {}

  if (Object.keys(rarites).length === 0) {
    throw new Error('No rarities rates received')
  }

  let rates = 0
  for (const rarity of Object.values(rarites)) {
    rates += rarity.rate
  }

  if (parseFloat(rates.toFixed(3)) !== 1) {
    const errorMessage = `Rarities rates sum must be equal to 1, ${rates} found`
    throw new Error(errorMessage)
  }

  if ((options?.traits || []).length === 0) {
    throw new Error('No traits received')
  }
}

function validateAssets(assets, options) {
  const availableTraits = options.traits.reduce((prev, current) => {
    prev[current] = true
    return prev
  }, {})

  for (const asset of assets) {
    if (!availableTraits[asset.trait]) {
      throw new Error(
        `${asset.trait} trait is not a valid one, found in ${asset} item`
      )
    }
  }
}

function buildTraitsMap(traitMap, traits, assets) {
  for (const trait of traits) {
    traitMap.set(trait, new Trait(trait))
  }

  for (const asset of assets) {
    traitMap.get(asset.trait).addElement(asset)
  }
}

function generateUnique(combinator, collection) {
  let generated = combinator.generate()

  while (collection.has(generated)) {
    generated = combinator.generate()
  }

  collection.add(generated)
}

module.exports = (options) => {
  const { rarities, traits } = options
  let combinator
  const collection = Collection()
  return {
    validateOptions: () => validateOptions(options),
    generateCollection: (assets, supply) => {
      const traitMap = collection.getTraitMap()

      validateAssets(assets, options)
      buildTraitsMap(traitMap, traits, assets)

      let maxSupply = 1

      for (const trait of traitMap.values()) {
        maxSupply *= trait.size()
      }

      if (maxSupply < supply) {
        throw new Error(
          `Requested supply ${supply} exceed maximum combinations ${maxSupply}`
        )
      }
      combinator = combinatorFactory(rarities, traitMap)

      for (let i = 0; i < supply; i++) {
        generateUnique(combinator, collection)
      }

      return collection
    },
  }
}
